/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


/*!
 *  \file main.c
 *  \brief oWL Pico reference design software.
 *
 *  Copyright (C) 2012 HD Wireless AB
 *
 *  You should have received a copy of the license along with this library.
 */


/*! \defgroup wlp_ref oWL Pico Reference Design
 *
 * The pclient application is built on top of the oWL pico API and uses the
 * owl toolkit, both provided by H&D Wireless.
 *
 * pclient provides a command line interface to directly invoke most of the
 * API functions in oWL Pico API, a ttcp throughput test tool and a small
 * web server.
 *
 * pclient can be built and used on a linux desktop PC together with the
 * SPB800 device. It can also be built for an ATMEL Xplain board, running on an
 * 8-bit CPU with 8Kb RAM. All the platform differences are handled in the
 * owl driver layer and in the main.c file.
 *
 * The pclient reference design can be used as a starting point for using the
 * oWL Pico API for a custom application and/or on a custom platform.
 * 
 *  @{
 */
#define OWL_TRACE OWL_TRACE_INFO

#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_lwip.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_led.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/components/spb8xx.h>

#include <wlp_api.h>
#include <wlp_spb8xx.h>
#include <wlp_inet.h>

#include <pico/wlp_shell.h>

void *owluart = NULL;

#define OWL_BOARD_SPB800P_UART ((void *) fd)

int 
uart_exchange(void *ctx, const void *tx, void *rx, int len)
{
        if (tx)
                return owluart_write(ctx, tx, len);
        if (rx)
                return owluart_read(ctx, rx, len);
}

static void
link_cb(void *ctx, int link)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "link_cb link:%d\n\r", link);

        if (link > 0)
                owlip_linkup();
        else 
                owlip_linkdown();
}

static void
addr_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        if (ip == NULL) {
                owlsh_printf(sh, "addr_cb ip:none\n\r");
        } else {
                owlsh_printf(sh, "addr_cb ip:%s\n\r", ip_ntoa(ip));
        }
}


void
systick_irq(void)
{
        owltmr_tick(OWL_BOARD_SYSTICK_PERIOD);
        debug_time += OWL_BOARD_SYSTICK_PERIOD;
}

static void
select_cb(void *bus, int select)
{
        return;
}

int
main(int argc, char *argv[])
{
        int err;
        struct owlsh_ctx *sh;
        int shflags = OWLSH_ESCAPECHARS;
        uint8_t fw_caps = 0;
        uint8_t buf[64];
        
        int fd;
        char *devname;

        owl_os_init();
        
	shflags |= OWLSH_NOECHO;
        
        if (argc > 1 && !owl_os_strcmp(argv[1], "-h")) {
                printf("%s [port]\n", argv[0]);
                return 0;
        }

        if(argc > 1)
                devname = argv[1];
        else
                devname = "/dev/ttyUSB0";


        if ((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
                owl_err("Failed to open device '%s'", devname);
                return -1;
        }

        owlboard_init();
	
        if ((err = owltmr_init()) < 0) {
                owl_err("owltmr_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

        /* Initialize the console UART */
        if (owluart_init(OWL_BOARD_SHELL_UART, UART_BAUDRATE_115200,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
                return -1;
        
        /* Initialize the host console input/output system. We will
         * only use the console for status output in the example
         * (using owlsh_print()) so the console command handling will
         * be unused.
         */
        if ((sh = owlsh_create(NULL, owluart_write, OWL_BOARD_SHELL_UART,
                               shflags)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }
        
        /* Power on SPB800 */
        spb8xx_resume();

        /* Initialize the host-WiFi UART */
        if ((err = owluart_init(OWL_BOARD_SPB800P_UART, UART_BAUDRATE_115200,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                owl_err("uart_init pico err:%d(%s)", err, owl_error(err));
                return -1;
        }
      
        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(
                     WLP_INIT_UART, uart_exchange, NULL, NULL,
                     OWL_BOARD_SPB800P_UART)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

        if (owlip_init()) {
                owlsh_printf(sh, "net init failed");
                return;
        }

        /* keep track of the uart fd to support the set_baudrate command */
        owluart = OWL_BOARD_SPB800P_UART;
      
        /* Get the firmware capabilities, so we know if the host attention pin
           is supported by the device or not */
        wlp_get_fw_caps(&fw_caps);
	
        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_link_cb(link_cb, sh);
        wlp_set_ipaddr_cb(addr_cb, sh);
        
        /* Add commands which can be used to control the WiFi link.
         * link_cb() will be called when the connection is up.
         */
        owlsh_addcmd(sh, &cmd_scan, NULL);
        owlsh_addcmd(sh, &cmd_set_wps, NULL);
        owlsh_addcmd(sh, &cmd_linkup, NULL);
        owlsh_addcmd(sh, &cmd_linkdown, NULL);
        owlsh_addcmd(sh, &cmd_get_hwaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_network, NULL);
        owlsh_addcmd(sh, &cmd_get_linkctx, buf);
        owlsh_addcmd(sh, &cmd_set_linkctx, buf);
        owlsh_addcmd(sh, &cmd_set_mode, NULL);
        owlsh_addcmd(sh, &cmd_get_mode, NULL);
        owlsh_addcmd(sh, &cmd_set_ps, NULL);
        owlsh_addcmd(sh, &cmd_conf_ps, NULL);
        owlsh_addcmd(sh, &cmd_set_channel, NULL);

        /* Add commands which can be used to control the ip address settings */
        owlsh_addcmd(sh, &cmd_set_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_get_dhcp, NULL);

        /* Add misc commands */
        owlsh_addcmd(sh, &cmd_get_version, NULL);
        owlsh_addcmd(sh, &cmd_get_fw_version, NULL);
        owlsh_addcmd(sh, &cmd_get_fw_caps, NULL);
        owlsh_addcmd(sh, &cmd_set_baudrate, NULL);
        owlsh_addcmd(sh, &cmd_set_led, NULL);
        owlsh_addcmd(sh, &cmd_set_heartbeat, NULL);
        owlsh_addcmd(sh, &cmd_set_poll_period, NULL);

        owlsh_addcmd(sh, &cmd_help, NULL);
 
        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */
        for (;;) {
                if (owluart_poll(OWL_BOARD_SHELL_UART))
                        owlsh_key(sh, owluart_inb(OWL_BOARD_SHELL_UART, 0));
     
                wlp_poll();
                owltmr_poll();
        }                      
        return 0;
}

/*!  @} */
