/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


/*!
 *  \file main.c
 *  \brief oWL Pico simple reference design software.
 *
 *  Copyright (C) 2012 HD Wireless AB
 *
 *  You should have received a copy of the license along with this library.
 */

/*
 * Minimal example for Wifi-connectivity using SPB8xx. This example
 * compiles for Linux. 
 *
 */

#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

/****************************************************************************
 * Platform-dependent code starts here. The functions, defines and includes
 * in this section must be adapted to the specific platform used.
 *
 *
 ****************************************************************************
 */

/*
 * Busy wait for the specified number of ms
 *
 * NOTE: This function must be modified for the specific platform used.
 *
 */
void
board_delay(int ms)
{
        while (ms--)
                usleep(1000);
}

/* 
 *
 * NOTE: This function must be modified for the specific platform used.
 *
 */
int
board_init(char *devname)
{
        int err;
        int spd;
        int fd;

        struct termios ti;
    
        if ((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
                printf("Failed to open device '%s'", devname);
                return -1;
        }
        
        /* Init uart: baudrate:115200, databits:8, parity:none, stopbits:1 */
        
        if ((err = tcgetattr(fd, &ti)) < 0) {
                printf("Failed to tcgetattr:%d", err);
                return -1;
        }

        cfmakeraw(&ti);
        ti.c_oflag = 0;

        spd = B115200;
        
        if ((err = cfsetispeed(&ti, spd)) < 0) {
                printf("cfsetispeed:%d", err);
                return -1;
        }
        
        if ((err = cfsetospeed(&ti, spd)) < 0) {
                printf("cfsetospeed:%d", err);
                return -1;
        }
	
        if ((err = tcsetattr(fd, TCSANOW, &ti)) < 0) {
                printf("Failed to tcsetattr:%d", err);
                return -1;
        }

        return fd;
}

/*
 * Write/Read len bytes to the UART. 
 * This function should block until len bytes have been written/read.
 *
 * NOTE: This function must be modified for the specific platform used.
 *
 */
int 
board_exchange(void *ctx, const void *tx, void *rx, int len)
{
        int pos = 0;
        int err;
        int fd = (int) ctx;

        if (tx) {
               
                if ((err = write(fd, tx, len)) != len) {
                        printf("write failed fd:%d len:%d err:%d", 
                               fd, len, err);
                        return -1;
                }
        }
        if (rx) {
                while (pos < len) {
                        if ((err = read(fd, rx + pos, len - pos)) < 0)
                                printf("read failed (fd:%d len:%d err:%d)",
                                       fd, len, err);
                        else 
                                pos += err;
                }
        }
        return len;
}

/****************************************************************************
 * Platform independent code starts here.
 * 
 *
 *
 ****************************************************************************
 */

#include <wlp_api.h>
#include <wlp_spb8xx.h>
#include <wlp_inet.h>

static void link_cb(void *ctx, int link)
{
        (void) ctx;
        printf("link_cb link:%d\n\r", link);
}


static void addr_cb(void *ctx, const struct ip_addr *ip)
{
        (void) ctx;

        if (ip == NULL) {
                printf("addr_cb ip:none\n\r");
        } else {
                printf("addr_cb ip:%s\n\r", ip_ntoa(ip));
        }
}

static void select_cb(void *uart, int select)
{
        return;
}

int main(int argc, char *argv[])
{
        int err;
        int uart;
        char *devname;

        struct wlp_ssid ssid;
        if (argc > 1 && !strcmp(argv[1], "-h")) {
                printf("%s [port]\n", argv[0]);
                return 0;
        }
        
        if(argc > 1)
                devname = argv[1];
        else
                devname = "/dev/ttyUSB0";
        
        if ((uart = board_init(devname)) < 0) {
                printf("board init failed");
                return -1;
        }

        /* Wait for SPB8XX to startup */
        board_delay(1300);
        
        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(
                     WLP_INIT_UART, 
                     board_exchange, 
                     NULL,
                     NULL,
                     (void *) uart)) < 0) {
                printf("wlp_init failed err:%d", err);
                return -1;
        }
	
        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_link_cb(link_cb, NULL);
        wlp_set_ipaddr_cb(addr_cb, NULL);

        /* specifiy the ssid ... */
        strcpy(ssid.ssid, "hd");

        /* ... and the number of valid bytes in the ssid */
        ssid.len = strlen("hd"); 
    
        /* start linkup process */
        if ((err = wlp_linkup(&ssid, 0, WLP_AUTH_AUTO)) < 0)
                printf("linkup failed err:%d", err);

        /* configure the device to use dhcp */ 
        if ((err = wlp_set_dhcp(1)) < 0)
                printf("set_dhcp failed err;%d", err);
        
        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */
        for (;;) 
                wlp_poll();
                                      
        return 0;
}


/*!  @} */
