/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 *
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_DATA
#define OWL_FILE "wlp_file"


#include <owl/core/owl_defs.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>

#include <wlp_socket.h>

static void send_file_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void recv_file_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static void write_fd(int fd, const char *buf, int len);
static void recv_cb(void *ctx, int sockid, int len);
static void conn_cb(void *ctx, int sockid, int connected);

static int parse_int(const char *str, int *out);

const struct owlsh_cmd cmd_send_file = {
        "send_file", send_file_f, "send_file <sockid> <file>"
};
const struct owlsh_cmd cmd_recv_file = {
        "recv_file", recv_file_f, "recv_file <sockid> <file>"
};

static int recvfd = -1;
static int recvid = WLP_INVALID_SOCKET_ID;

static void
send_file_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int sockid;
        const char *name;
        char buf[256];
        int read_err = 0;
        int send_err = 0;
        int len = 0;
        int fd;
        struct stat stat_buf;
        int file_len;
        int read_len = 0;

        if (argc < 3)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        name = argv[2];

        if ((fd = open(name, O_RDONLY)) < 0) {
                owlsh_printf(sh, "could not open '%s' fd:%d\r\n", name, fd);
                return;
        }

        if (stat(name, &stat_buf) < 0) {
                close(fd);
                owlsh_printf(sh, "could not stat %s\r\n", name);
                return;
        }

        file_len = stat_buf.st_size;
        owlsh_printf(sh, "sending %d bytes\r\n", file_len);

        while (read_len < file_len) {
                if ((read_err = read(fd, buf, sizeof(buf))) < 0)
                        goto out;

                read_len += read_err;
                send_err = 0;
                int buflen = read_err;
                int bufpos = 0;
                while (bufpos < buflen) {
                        if ((send_err = wlp_send(sockid, buf + bufpos,
                                                 buflen - bufpos)) < 0)
                                goto out;
                        len += send_err;
                        bufpos += send_err;
                        owlsh_printf(sh, "wlp_send: %d\r\n", send_err);
                }
        }

out:
        owlsh_printf(sh, "done -- read_err:%d send_err:%d len:%d\r\n",
                     read_err, send_err, len);
        close(fd);
        return;
err_parse:
        owlsh_printf(sh, "usage: send_file <sockid> <file>\r\n");
}

static void
recv_file_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int backlog = 1;
        const char *name;
        int sockid;

        if (argc < 3)
                goto err_parse;
        if (!parse_int(argv[1], &sockid))
                goto err_parse;
        name = argv[2];

        if (recvfd >= 0) {
                owlsh_printf(sh, "closing fd:%d\r\n", recvfd);
                close(recvfd);
        }

        if ((recvfd = open(name, O_CREAT | O_RDWR | O_TRUNC, 0660)) < 0) {
                owlsh_printf(sh, "failed to open %s\r\n", name);
                return;
        }

        wlp_set_conn_cb(sockid, conn_cb, sh);
        wlp_set_recv_cb(sockid, recv_cb, sh);
 
        owlsh_printf(sh, "rx data will be written to file:%s\r\n", name);
        return;


err_parse:
        owlsh_printf(sh, "usage: recv_file <sockid> <file>\r\n");
}

static void
recv_cb(void *ctx, int sockid, int len)
{
        int res;
        struct owlsh_ctx *sh = ctx;
        char buf[64];

        
        owlsh_printf(sh, "recv_cb sockid:%d len:%d %s\r\n", sockid, len,
                     len ? "" : "(connection lost)");

        while (len > 0) {
                int recvlen;
                if ((uint32_t) len > sizeof(buf))
                        recvlen = sizeof(buf);
                else
                        recvlen = len;
                
                res = wlp_recv(sockid, buf, recvlen);
                if (res < 0) {
                        owlsh_printf(sh, "res < 0\r\n");
                        owl_err("fail to recv %d bytes err:%d\r\n",
                                recvlen, res);
                }
                else if (res != len) {
                        owlsh_printf(sh, "res != len\r\n");
                        owl_info("got %d bytes, %d remains\r\n",
                                 res, len - res);
                }

                if (res > 0) {
                        owlsh_printf(sh, "res > 0\r\n");
                        len -= res;
                        write_fd(recvfd, buf, res);
                }
        }
        
}

static void
conn_cb(void *ctx, int sockid, int connected)
{
        struct owlsh_ctx *sh = ctx;

        if (sockid == recvid) {
                owlsh_printf(sh, "closing %d\r\n", recvfd);
                close(recvfd);
                sockid = WLP_INVALID_SOCKET_ID;
        }
        
        owlsh_printf(sh, "conn_cb sockid:%d connected:%d\r\n",
                     sockid, connected);
}

static void
write_fd(int fd, const char *buf, int len)
{
        int fpos = 0;
        int err;
        while (fpos < len) {
                if ((err = write(fd, buf + fpos, len - fpos)) < 0) {
                        owl_err("write fail:%d", err);
                        return;
                }
                fpos += err;
        }
}

static int
parse_int(const char *str, int *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}
