/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */


/*!
 *  \file main.c
 *  \brief oWL Pico reference design software.
 *
 *  Copyright (C) 2012 HD Wireless AB
 *
 *  You should have received a copy of the license along with this library.
 */


/*! \defgroup wlp_ref oWL Pico Reference Design
 *
 * The pclient application is built on top of the oWL pico API and uses the
 * owl toolkit, both provided by H&D Wireless.
 *
 * pclient provides a command line interface to directly invoke most of the
 * API functions in oWL Pico API, a ttcp throughput test tool and a small
 * web server.
 *
 * pclient can be built and used on a linux desktop PC together with the
 * SPB800 device. It can also be built for an ATMEL Xplain board, running on an
 * 8-bit CPU with 8Kb RAM. All the platform differences are handled in the
 * owl driver layer and in the main.c file.
 *
 * The pclient reference design can be used as a starting point for using the
 * oWL Pico API for a custom application and/or on a custom platform.
 * 
 *  @{
 */
#define OWL_TRACE OWL_TRACE_DBG

#include <errno.h>
#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_rtc.h>
#include <owl/drivers/owl_led.h>
#include <owl/drivers/owl_spi.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/drivers/owl_sdio.h>
#include <owl/components/spb8xx.h>

#include <wlp_api.h>
#include <wlp_spb8xx.h>
#include <wlp_inet.h>

#include <pico/wlp_shell.h>
#include <pico/wlp_ttcp.h>
#include <pico/wlp_ntp.h>
#include <pico/wlp_ping.h>
#include <pico/wlp_httpd.h>

#include "board.h"

extern uint32_t __get_MSP(void);

void *owluart = NULL;

#define TIMER_PERIOD_MS 64

#ifdef __GNUC__
/**
 * sbrk
 *
 * Increase program data space. Malloc and related functions depend on this.
 */
char *_sbrk(int incr)
{
        extern char _ebss; /* Defined by the linker */
        static char *heap_end;
        char *prev_heap_end;

        if (heap_end == 0)
                heap_end = &_ebss;

        prev_heap_end = heap_end;

        char *stack = (char *) __get_MSP();
        if (heap_end + incr >  stack) {
                errno = ENOMEM;
                return  (char *) -1;
        }

        heap_end += incr;
        return (char *) prev_heap_end;
}
#endif /* __GNUC__ */

#if defined(OWL_BOARD_SPB8XX_UART)
int 
uart_exchange(void *ctx, const void *tx, void *rx, int len)
{
        int err = 0;

        if (tx) {
                err = owluart_write(ctx, tx, len);
        }
	if (rx) {
		while ((err = owluart_poll(ctx)) < len);
		err = owluart_read(ctx, rx, len);
	}
	return err;
}
#endif
               

static void link_cb(void *ctx, int link)
{
        struct owlsh_ctx *sh = ctx;
        
	owlsh_printf(sh, "link_cb link:%d\n\r", link);
}


static void addr_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        
	if (ip == NULL) {
                owlsh_printf(sh, "addr_cb ip:none\n\r");
        } else {
                owlsh_printf(sh, "addr_cb ip:%s\n\r", ip_ntoa(ip));
        }
}

#if defined(OWL_BOARD_SPB8XX_SPI)
static void select_cb(void *ctx, int select)
{       
        owl_unused(ctx);
        if (select)
                owlspi_claim(OWL_BOARD_SPB8XX_SPI, 
                             OWL_BOARD_SPB8XX_SPI_CS);
        else
                owlspi_release(OWL_BOARD_SPB8XX_SPI, 
                               OWL_BOARD_SPB8XX_SPI_CS);
}
#endif /* OWL_BOARD_SPB8XX_SPI */

#if defined(OWL_BOARD_SPB8XX_HOST_ATTN_PIN)
int ha_pin_is_signalled(void)
{
        if (owlgpio_is_low(OWL_BOARD_SPB8XX_HOST_ATTN_PIN))
                return 1;
        
        return 0;
}
#endif /* OWL_BOARD_SPB8XX_HOST_ATTN_PIN */

int main(int argc, char *argv[])
{
        owl_unused(argc);
        owl_unused(argv);

        int err;
        uint8_t mode = 0;
        uint8_t fw_caps = 0;
	struct owlsh_ctx *sh;
	/* Buffer to store linkctx */
	uint8_t buf[256];

        owl_unused(argc);
        owl_unused(argv);
      	
	owl_os_init();

        owlboard_init();
  
#if defined(OWL_BOARD_SHELL_UART)
        /* Initialize the console UART */
        if (owluart_init(OWL_BOARD_SHELL_UART, UART_BAUDRATE_115200,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
                return -1;

	/* Initialize the host console input/output system. We will
         * only use the console for status output in the example
         * (using owlsh_print()) so the console command handling will
         * be unused.
         */
        if ((sh = owlsh_create(NULL, 
			       owluart_write, 
			       OWL_BOARD_SHELL_UART,
                               0)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }

#endif
#if defined(OWL_BOARD_SPB8XX_UART)
	/* Initialize the SPB8XX UART */
        if (owluart_init(OWL_BOARD_SPB8XX_UART, UART_BAUDRATE_115200,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
	        return -1; 

	if (owluart_set_bufsize(OWL_BOARD_SPB8XX_UART, 512) < 0)
		return -1;
#endif
#if defined(OWL_BOARD_SPB8XX_HOST_ATTN_PIN)
        owlgpio_cfg(OWL_BOARD_SPB8XX_HOST_ATTN_PIN, OWL_GPIO_INPUT);
#endif /* OWL_BOARD_SPB8XX_HOST_ATTN_PIN */

	if ((err = owltmr_init()) < 0) {
                owl_err("owltmr_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

	/* Power on SPB800 */
        spb8xx_resume();

#if defined(OWL_BOARD_SPB8XX_UART)
        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(WLP_INIT_UART,
                            uart_exchange,
                            NULL,
                            NULL,
                            OWL_BOARD_SPB8XX_UART)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

        /* keep track of the uart to support the set_baudrate command */
        owluart = OWL_BOARD_SPB8XX_UART;

#elif defined(OWL_BOARD_SPB8XX_SPI)
        /* Initialize the host-WiFi SPI bus */
        if ((err = owlspi_init(OWL_BOARD_SPB8XX_SPI,
                               OWL_SPI_MODE_MASTER)) < 0) {
                owl_err("spi err:%d(%s)", err, owl_error(err));
                return -1;
        }

        if ((err = owlspi_setup_cs(OWL_BOARD_SPB8XX_SPI,
                                   OWL_BOARD_SPB8XX_SPI_CS,
                                   OWL_BOARD_SPB8XX_SPI_FREQ,
                                   OWL_SPI_TRANSFER_MODE_3, 8)) < 0) {
                owl_err("spi err:%d(%s)", err, owl_error(err));
                return -1;
        }

        if ((err = wlp_init(WLP_INIT_SPI, 
                            owlspi_exchange,
                            select_cb, 
                            NULL,
                            OWL_BOARD_SPB8XX_SPI) < 0)) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

#elif defined(OWL_BOARD_SPB8XX_SDIO) 
        /* Initialize the host-WiFi SDIO bus */

        int flags = OWL_SDIO_FLAG_4BIT | OWL_SDIO_FLAG_DMA;

        if ((err = owlsdio_init(NULL,
                                &flags,
                                OWL_SDIO_MODE_MASTER)) < 0) {
                owl_err("spi err:%d(%s)", err, owl_error(err));
                return -1;
        }

        if ((err = wlp_init(WLP_INIT_SDIO, 
                            owlsdio_exchange,
                            NULL, 
                            owlsdio_cmd,
                            NULL) < 0)) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

#endif /* OWL_BOARD_SPB8XX_UART */

#if defined(OWL_BOARD_SPB8XX_WAKEUP_PIN)
        /* Register a wakeup pin control callback.
         * This is used to wake the SPB800 when Power save level 1 is
         * used. */
        wlp_set_wakeup_cb(wakeup_cb, sh);
#endif /* defined(OWL_BOARD_SPB8XX_WAKEUP_PIN) */

        /* Get the firmware capabilities, so we know if the host 
           attention pin is supported by the device or not */
        wlp_get_fw_caps(&fw_caps);

        /* Get the current operating mode */
        wlp_get_mode(&mode);
	
        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_link_cb(link_cb, sh);
        wlp_set_ipaddr_cb(addr_cb, sh);

#if defined(OWL_BOARD_SHELL_UART)
	/* Add commands which can be used to control the WiFi link.
         * link_cb() will be called when the connection is up.
         */
	owlsh_addcmd(sh, &cmd_scan, NULL);
        owlsh_addcmd(sh, &cmd_set_wps, NULL);
        owlsh_addcmd(sh, &cmd_linkup, NULL);
        owlsh_addcmd(sh, &cmd_linkdown, NULL);
        owlsh_addcmd(sh, &cmd_get_hwaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_network, NULL);
        owlsh_addcmd(sh, &cmd_get_linkctx, buf);
        owlsh_addcmd(sh, &cmd_set_linkctx, buf);

        /* Add commands which can be used to control the ip address settings */
        owlsh_addcmd(sh, &cmd_set_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_get_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcpd, NULL);
        owlsh_addcmd(sh, &cmd_get_hostbyname, NULL);

        /* Add commands which can be used to play around with the socket
         * interface of owl pico
         */
        owlsh_addcmd(sh, &cmd_socket, NULL);
        owlsh_addcmd(sh, &cmd_bind, NULL);
        owlsh_addcmd(sh, &cmd_listen, NULL);
        owlsh_addcmd(sh, &cmd_connect, NULL);
        owlsh_addcmd(sh, &cmd_close, NULL);
        owlsh_addcmd(sh, &cmd_send, NULL);
        owlsh_addcmd(sh, &cmd_sendto, NULL);
        owlsh_addcmd(sh, &cmd_get_peeraddr, NULL);

        /* Add misc commands */
        owlsh_addcmd(sh, &cmd_get_version, NULL);
        owlsh_addcmd(sh, &cmd_get_fw_version, NULL);
        owlsh_addcmd(sh, &cmd_get_fw_caps, NULL);
        owlsh_addcmd(sh, &cmd_set_mode, NULL);
        owlsh_addcmd(sh, &cmd_get_mode, NULL);
        owlsh_addcmd(sh, &cmd_set_ps, NULL);
        owlsh_addcmd(sh, &cmd_conf_ps, NULL);
        owlsh_addcmd(sh, &cmd_set_channel, NULL);
        owlsh_addcmd(sh, &cmd_set_baudrate, NULL);
        owlsh_addcmd(sh, &cmd_fw_upgrade, NULL);
        owlsh_addcmd(sh, &cmd_set_led, NULL);
        owlsh_addcmd(sh, &cmd_set_heartbeat, NULL);
        owlsh_addcmd(sh, &cmd_set_poll_period, NULL);
        owlsh_addcmd(sh, &cmd_loopback, NULL);

        /* Add protocol implementation examples */
        owlsh_addcmd(sh, &cmd_ttcp, NULL);
        owlsh_addcmd(sh, &cmd_ntp, NULL);
        owlsh_addcmd(sh, &cmd_ping, NULL);
	owlsh_addcmd(sh, &cmd_help, NULL);

#if OWL_CONFIG_MEMTRACE
        owlsh_addcmd(sh, &cmd_memtrace, NULL);
#endif /* OWL_CONFIG_MEMTRACE */
#endif
        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */
        for (;;) {
#if defined(OWL_BOARD_SHELL_UART)
	        if (owluart_poll(OWL_BOARD_SHELL_UART))
		        owlsh_key(sh, owluart_inb(OWL_BOARD_SHELL_UART, 0));
#endif
                /* If the host has support for the host attention pin, we will
                   use it if also supported by the device */
#if defined(OWL_BOARD_SPB8XX_HOST_ATTN_PIN)
                if (fw_caps & WLP_CAP_HA_PIN) {
                        /* If the host attention pin is present we
                         * only poll if the pin is asserted (active low).
                         * When the pin is asserted there's an event
                         * pending from the SPB800 module. */
	      
                        if (ha_pin_is_signalled()) {
                                owl_vdbg("poll");
                                wlp_poll();
                        }
			
                } else {
                        /* Without the host attention pin we need to
                         * poll the SPB800 periodically. This method will
                         * prevent the SPB800 from using power save
                         * level 1. */ 
                        wlp_poll();
                }
#else 
                wlp_poll();
#endif
                owltmr_poll();
        }                      
        return 0;
}

void EXTI9_5_IRQHandler(void)
{       
#if defined(OWL_BOARD_SPB8XX_SPI_IRQ_PIN)
        if (EXTI_GetITStatus(EXTI_Line9) != RESET) {
                wlp_irq();
                
                /* Clear the  EXTI line 13 pending bit */
                EXTI_ClearITPendingBit(EXTI_Line9);
        }
#endif
}

void SDIO_IRQHandler(void)
{
        SDIO_ClearITPendingBit(SDIO_IT_SDIOIT);
        wlp_irq();
}

void
SysTick_Handler(void)
{
        owltmr_tick(OWL_BOARD_SYSTICK_PERIOD);
}

/*!  @} */
