/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR

#include <owl/core/owl_debug.h>
#include <owl/core/owl_defs.h>
#include <owl/core/owl_err.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_gpio.h>
#include <owl/drivers/owl_uart.h>
#include <owl/components/spb8xx.h>

#include <wlp_spb8xx.h>
#include <wlp_ipaddr.h>
#include <wlp_inet.h>

static void set_baudrate_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_led_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_heartbeat_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void set_poll_period_f(struct owlsh_ctx *sh, int argc,
                              const char *argv[]);
static void get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void get_fw_version_f(struct owlsh_ctx *sh, int argc,
                             const char *argv[]);
static void get_fw_caps_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void fw_upgrade_f(struct owlsh_ctx *sh, int argc, const char *argv[]);
static void loopback_f(struct owlsh_ctx *sh, int argc, const char *argv[]);

static int parse_int(const char *str, int *out);
static int32_t parse_int32(const char *str, int32_t *out);

const struct owlsh_cmd cmd_set_baudrate = {
        "set_baudrate", set_baudrate_f, "wlp_set_baudrate()"
};
const struct owlsh_cmd cmd_get_fw_caps = {
        "get_fw_caps", get_fw_caps_f, "print firmware capabilities"
};
const struct owlsh_cmd cmd_set_led = {
        "set_led", set_led_f, "wlp_set_led()"
};
const struct owlsh_cmd cmd_set_heartbeat = {
        "set_heartbeat", set_heartbeat_f, "wlp_set_heartbeat()"
};
const struct owlsh_cmd cmd_set_poll_period = {
        "set_poll_period", set_poll_period_f, "wlp_set_poll_period()",
};
const struct owlsh_cmd cmd_get_version = {
        "get_version", get_version_f, "print client version"
};
const struct owlsh_cmd cmd_get_fw_version = {
        "get_fw_version", get_fw_version_f, "print firmware version"
};
const struct owlsh_cmd cmd_fw_upgrade = {
        "fw_upgrade", fw_upgrade_f, "wlp_fw_upgrade(method, addr, port, path)"
};
const struct owlsh_cmd cmd_loopback = {
        "loopback", loopback_f, "wlp_loopback()"
};

int
spb8xx_resume(void)
{
        /* Host attention should be an input */
#if defined(OWL_BOARD_SPB8XX_HOST_ATTENTION_PIN)
        owl_dbg("conf host attention pin as input");
        owlgpio_cfg(OWL_BOARD_SPB8XX_HOST_ATTENTION_PIN, OWL_GPIO_INPUT);
#endif

        /* Wakeup should be an output. 
	   At spb8xx startup we need to set the transport mode (uart, spi)
        */
#if defined(OWL_BOARD_SPB8XX_WAKEUP_SEL_PIN)
        owl_dbg("conf wakeup pin to wakeup SPB8XX");
        owlgpio_cfg(OWL_BOARD_SPB8XX_WAKEUP_PIN, OWL_GPIO_OUTPUT);
#if defined(OWL_BOARD_SPB8XX_UART)       
	owlgpio_clr(OWL_BOARD_SPB8XX_WAKEUP_PIN);
#elif defined(OWL_BOARD_SPB8XX_SPI)
	owlgpio_set(OWL_BOARD_SPB8XX_WAKEUP_PIN);
#endif /* OWL_BOARD_SPB8XX_UART */
#endif /* OWL_BOARD_SPB8XX_WAKEUP_PIN */

        /* Now we plug the power to SPB800 */
#if defined (OWL_BOARD_SPB8XX_SHUTDOWN_PIN)
        owl_dbg("conf shutdown pin to power on SPB8XX");
        owlgpio_cfg(OWL_BOARD_SPB8XX_SHUTDOWN_PIN, OWL_GPIO_OUTPUT);
        owlgpio_set(OWL_BOARD_SPB8XX_SHUTDOWN_PIN);
        owlboard_mdelay(500);
        owlgpio_clr(OWL_BOARD_SPB8XX_SHUTDOWN_PIN);
#endif

	/* Set initial state of wakeup-pin after being used as
	   spi/uart mode select */
#if defined(OWL_BOARD_SPB8XX_WAKEUP_PIN)
	owlgpio_set(OWL_BOARD_SPB8XX_WAKEUP_PIN);
#endif /* OWL_BOARD_SPB8XX_WAKEUP_PIN */

#if defined(OWL_BOARD_SPB8XX_HOST_ATTN_PIN)
	while (owlgpio_is_high(OWL_BOARD_SPB8XX_HOST_ATTN_PIN));
        owl_info("wait for IRQ to go high");
        while (owlgpio_is_low(OWL_BOARD_SPB8XX_HOST_ATTN_PIN));
        owl_info("device started");
	owlboard_mdelay(150);
#else
	/* Wait for wifi to start */
        owl_dbg("wait for SPB8XX to start");
        owlboard_mdelay(1300);
#endif
        
        return 0;
}

int
spb8xx_suspend(void)
{
#if defined (OWL_BOARD_SPB8XX_SHUTDOWN_PIN)
        owl_dbg("conf shutdown pin to power off SPB8XX");
        owlgpio_cfg(OWL_BOARD_SPB8XX_SHUTDOWN_PIN, OWL_GPIO_OUTPUT);
        owlgpio_set(OWL_BOARD_SPB8XX_SHUTDOWN_PIN);
#endif
        return 0;
}


static void
set_baudrate_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int err;
        int32_t baudrate = 0;
        int rtscts = 0;
        extern void *owluart; /* defined in main.c */

        if (argc >= 3) {
                /* set_baudrate <baudrate> <rtscts> */
                if (!parse_int32(argv[1], &baudrate))
                        goto err_parse;
                if (!parse_int(argv[2], &rtscts))
                        goto err_parse;

        } else if (argc >= 2) {
                /* set_baudrate <baudrate> */
                if (!parse_int32(argv[1], &baudrate))
                        goto err_parse;

        } else {
                goto err_parse;
        }

        res = wlp_set_baudrate(baudrate, rtscts);
        if (!res) {
                owlboard_mdelay(250);
                if ((err = owluart_init(owluart, baudrate,
                                        UART_DATABITS_8, UART_PARITY_NONE,
                                        UART_STOPBITS_1, rtscts)) < 0)
                        goto err_uart;
        }

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_baudrate <baudrate> [rtscts]\r\n");
        return;

err_uart:
        owlsh_printf(sh, "err %d (failed to configure host uart)\r\n", err);
}

static void
set_led_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int enable;

        if (argc < 2)
                goto err_parse;
        if (!parse_int(argv[1], &enable))
                goto err_parse;
        res = wlp_set_led(enable);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_led <0|1>\r\n");

}

static void
set_heartbeat_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int32_t period;

        if (argc < 2)
                goto err_parse;
        if (!parse_int32(argv[1], &period))
                goto err_parse;
        res = wlp_set_heartbeat((uint32_t)period);
        owlsh_printf(sh, "period %d\r\n", (int) period);
        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_heartbeat <0|1>\r\n");

}

static void
set_poll_period_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        int32_t period;

        if (argc < 2)
                goto err_parse;
        if (!parse_int32(argv[1], &period))
                goto err_parse;
        res = wlp_set_poll_period((uint32_t)period);

        owlsh_perr(sh, res);
        return;

err_parse:
        owlsh_printf(sh, "usage: set_poll_period <period in ms>\r\n");

}

static void
fw_upgrade_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        struct ip_addr addr;
        int res;
        uint16_t port = 8080;
        uint8_t method = WLP_UPGRADE_METHOD_HTTP_POST;
        char *path = NULL;

        wlp_get_ipaddr(&addr, NULL, NULL, NULL);

        if (argc > 5 || argc < 1)
                goto err_parse;

        if (argc > 1) {
                if (!owl_os_strcmp(argv[1], "http_post"))
                        method = WLP_UPGRADE_METHOD_HTTP_POST;
                else
                        goto err_parse;
        }

        if (argc > 2) {
                if (!ip_aton(argv[2], &addr))
                        goto err_parse;
        }

        if (argc >= 4)
                port = owl_os_atoi(argv[3]);

        if (argc >= 5)
                path = (char *)argv[4];

        res = wlp_fw_upgrade(method, &addr, port, path);
        owlsh_perr(sh, res);
        if (WLP_UPGRADE_METHOD_HTTP_POST == method)
                owlsh_printf(sh, "Listening on %s:%d/%s\r\n", ip_ntoa(&addr),
                             port, path ? path : "");
        return;

err_parse:
        owlsh_printf(sh, "usage: fw_upgrade <method, e.g. \"http_post\"> [<ip> <port> <path>]\r\n");

}

static void
get_version_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        (void)argc;
        (void)argv;

        owlsh_printf(sh, "sw version:%d.%d.%d\r\n",
                     WLP_MAJOR_VERSION(WLP_VERSION_CODE),
                     WLP_MINOR_VERSION(WLP_VERSION_CODE),
                     WLP_PATCH_VERSION(WLP_VERSION_CODE));
}

static void
get_fw_version_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint32_t code;
        char build[16];
        char hw_rev[16];

        (void)argc;
        (void)argv;

        res = wlp_get_fw_version(&code, build, sizeof build,
                                 hw_rev, sizeof hw_rev);
        owlsh_perr(sh, res);
        owlsh_printf(sh, "fw version:%d.%d.%d build:%s\r\n",
                     WLP_MAJOR_VERSION(code), WLP_MINOR_VERSION(code),
                     WLP_PATCH_VERSION(code), build);
        owlsh_printf(sh, "hw revision:%s\r\n",
                     owl_os_strlen(hw_rev) ? hw_rev : "unknown");
}

static void
get_fw_caps_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
        int res;
        uint8_t caps;

        (void)argc;
        (void)argv;

        res = wlp_get_fw_caps(&caps);
        owlsh_perr(sh, res);
        owlsh_printf(sh, "caps:0x%02x\r\n", caps);
}

static void
loopback_f(struct owlsh_ctx *sh, int argc, const char *argv[])
{
	int i;
	int err;
	int len;
	char *out; 
        char *in;
	const char *usage = "usage: loopback <len> [data]\r\n";
	
	if (argc < 3) {
		owlsh_printf(sh, "%s", usage);
		return;
	}
      
	len = owl_os_strtol(argv[1], NULL, 0);
	
	if ((out = owl_os_zalloc(len)) == NULL) {
		owl_err("could not alloc out-buf");
		owlsh_printf(sh, "out of mem");
		return;
	}

        if ((in = owl_os_zalloc(len)) == NULL) {
		owl_err("could not alloc in-buf");
		owlsh_printf(sh, "out of mem");
		return;
	}

	for (i = 0; i < len; i++) {
                int arg = i + 2;
                if (arg < argc) {
                        /* do we have a value on the command line? */
                        out[i] = owl_os_strtol(argv[arg], NULL, 0);
                } else {
			if (i > 0) 
			        out[i] = i;
                        else 
                                out[i] = 0;
                }
        }

	if ((err = wlp_loopback(in, out, len))) {
		owlsh_perr(sh,err);
	}

        for (i = 0; i < len; i++) {
                if (in[i] != out[i]) {
                        owl_err("wrong data on idx:%d", i);
                        owlsh_printf(sh, "loopback failed, idx:%d in:%x out:%x", i, in[i], out[i]);
                }
        }

	owl_os_free(in);
        owl_os_free(out);
}

static int
parse_int(const char *str, int *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}

static int32_t
parse_int32(const char *str, int32_t *out)
{
        char *end;
        *out = owl_os_strtol(str, &end, 0);
        if (end == str)
                return 0;

        return 1;
}
