/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#define OWL_TRACE OWL_TRACE_ERR
#define OWL_EVENTLOG OWL_EVENTLOG_DBG
#define OWL_FILE "owl_lwip"

#include <owl/core/owl_defs.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
#include <owl/core/owl_timer.h>
#include <owl/core/owl_eventlog.h>
#include <owl/core/owl_lwip.h>

#include <lwip/init.h>
#include <lwip/tcp_impl.h>
#include <lwip/ip_frag.h>
#if OWL_CONFIG_SOCKET_BSD
# include <lwip/tcpip.h>
#endif
#include <netif/etharp.h>
#include <netif/wlif.h>

static void ip_status_cb(struct netif *netif);
#if !OWL_CONFIG_SOCKET_BSD
static void tcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx);
static void dhcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx);
static void wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx);
#endif /* !OWL_CONFIG_SOCKET_BSD */

#define DHCP_ADDR_TMO (5 * 60000)

/* Declared in owl/core/owl_lwip.h */
struct owlip_ctx *owlip_ctx = NULL;

static struct owlip_ctx *priv = NULL;

int
owlip_init(void)
{
        struct ip_addr ipaddr, netmask, gw;

        owl_info("enter");

        if (priv != NULL) {
                owl_info("already initialized");
                wlif_reinit(priv->netif);
                return 0;
        }

        IP4_ADDR(&gw, 0,0,0,0);
        IP4_ADDR(&ipaddr, 0,0,0,0);
        IP4_ADDR(&netmask, 0,0,0,0);

        /* With OS 'owl', this is done in owl_os_init() */
#if !OWL_CONFIG_SOCKET_BSD
# if !OWL_CONFIG_OS_OWL
        lwip_init();
# endif
#else
        tcpip_init(NULL, NULL);
#endif /* !OWL_CONFIG_SOCKET_BSD */

        if ((priv = owl_os_zalloc(sizeof(*priv))) == NULL)
                return OWL_ERR_MEM;

        if ((priv->netif = owl_os_zalloc(sizeof(struct netif))) == NULL) {
                owl_os_free(priv);
                priv = NULL;
                return OWL_ERR_MEM;
        }

        
        if ((netif_add(
		     priv->netif, &ipaddr, &netmask, &gw, NULL, wlif_init,
#if !OWL_CONFIG_SOCKET_BSD
		     ethernet_input
#else
		     tcpip_input
#endif /* !OWL_CONFIG_SOCKET_BSD */
		     )) == NULL) {
                owl_os_free(priv->netif);
                owl_os_free(priv);
                priv = NULL;
                return OWL_FAILURE;
        }
        netif_set_default(priv->netif);
        netif_set_status_callback(priv->netif, ip_status_cb);

#if !OWL_CONFIG_SOCKET_BSD
        owltmr_set_name(&priv->tcp_tmo, "tcp");
        owltmr_set_name(&priv->dhcp_tmo, "dhcp");

        if (owl_bit(wlif_flags, WLIF_FLAG_POLL))
                owltmr_addtmo(0, 1, wifi_tmr_cb, NULL);
#endif /* !OWL_CONFIG_SOCKET_BSD */

        owlip_ctx = priv;
        return 0;
}

void
owlip_linkup(void)
{
        owl_eventstr_info("linkup");
        
        netif_set_link_up(priv->netif);
        
        if (owl_bit(priv->flags, FLAG_DHCP)) {

                if (priv->dhcp_state == DHCP_STATE_IDLE ||
                    priv->dhcp_state == DHCP_STATE_INFORMED) {
                        struct ip_addr ipaddr, netmask, gw;

#if OWL_CONFIG_HIBERNATE
                        if (owl_bit(priv->flags, FLAG_DHCP_INFORM)) {
                                netif_set_ipaddr(priv->netif, &priv->ipaddr);
                                netif_set_netmask(priv->netif, &priv->netmask);
                                netif_set_gw(priv->netif, &priv->gw);
                                dns_setserver(0, &priv->dns);

                                owl_info("linkup, informing dhcp");
                                dhcp_inform(priv->netif);
                                netif_set_up(priv->netif);
                                priv->dhcp_state = DHCP_STATE_INFORMED;
                        } else
#endif /* OWL_CONFIG_HIBERNATE */
                        {
                                IP4_ADDR(&gw, 0,0,0,0);
                                IP4_ADDR(&ipaddr, 0,0,0,0);
                                IP4_ADDR(&netmask, 0,0,0,0);
                                netif_set_ipaddr(priv->netif, &ipaddr);
                                netif_set_netmask(priv->netif, &netmask);
                                netif_set_gw(priv->netif, &gw);

                                owl_info("linkup, requesting dhcp");
                                dhcp_start(priv->netif);
                                if (!owltmr_is_sched(&priv->dhcp_tmo))
                                        owltmr_sched(
                                                &priv->dhcp_tmo,
                                                DHCP_COARSE_TIMER_MSECS,
                                                1, dhcp_tmr_cb, NULL);

                                priv->dhcp_state = DHCP_STATE_RUN;
                        }
                }

        } else if (!ip_addr_isany(&priv->netif->ip_addr)) {
                /* bring the interface up if a static ip is configured */

                owl_info("linkup, set interface up");
                netif_set_up(priv->netif);

        } else {
                owl_info("linkup, netif still down");
        }

        owlip_start();
        priv->linkup_time = owltmr_get_ms();
}

void
owlip_linkdown(void)
{
        owl_eventstr_info("linkdown");
        netif_set_link_down(priv->netif);
}

void
owlip_start(void)
{
#if !OWL_CONFIG_SOCKET_BSD
        owltmr_sched(&priv->tcp_tmo, TCP_TMR_INTERVAL, 1, tcp_tmr_cb, NULL);
#endif
}

void
owlip_stop(void)
{
#if !OWL_CONFIG_SOCKET_BSD
        owltmr_cancel(&priv->tcp_tmo);
#endif
}

/* Used to be lwip_error and lived in owl_net.c */
int
owlip_error(err_t err)
{
        switch (err) {
        case ERR_OK:
                return 0;
        case ERR_ISCONN: /* fall through */
        case ERR_RTE:
                return OWL_ERR_STATE;
        case ERR_VAL:
                return OWL_ERR_PARAM;
        case ERR_MEM:
                return OWL_ERR_MEM;
        case ERR_USE:
                return OWL_ERR_INUSE;
        case ERR_INPROGRESS:
                return OWL_ERR_BUSY;
        default:
                owl_err("unknown lwip return code:%d", err);
                return OWL_FAILURE;
        };
}

static void
ip_status_cb(struct netif *netif)
{
        if (netif_is_up(netif)) {
                owl_eventstr_info("ip:1");
                owl_info("bound to %d.%d.%d.%d",
                         ip4_addr1(&netif->ip_addr),
                         ip4_addr2(&netif->ip_addr),
                         ip4_addr3(&netif->ip_addr),
                         ip4_addr4(&netif->ip_addr));

#if OWL_CONFIG_HIBERNATE
                ip_addr_set(&priv->ipaddr, &netif->ip_addr);
                ip_addr_set(&priv->netmask, &netif->netmask);
                ip_addr_set(&priv->gw, &netif->gw);
                priv->dns = dns_getserver(0);
                owl_bitset(priv->flags, FLAG_DHCP_INFORM);
#endif /* OWL_CONFIG_HIBERNATE */

                if (priv->addr_cb)
                        priv->addr_cb(priv->addr_ctx, &netif->ip_addr);
                
        } else {
                owl_eventstr_info("ip:0");
                owl_info("ip status down");
                if (priv->addr_cb)
                        priv->addr_cb(priv->addr_ctx, NULL);
        }
}

#if !OWL_CONFIG_SOCKET_BSD
static void
tcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        owl_unused(tmo);
        owl_unused(ctx);

        priv->tick++;

        tcp_tmr();
        if (!(priv->tick % (5000 / TCP_TMR_INTERVAL)))
                etharp_tmr();
        if (!(priv->tick % (DHCP_FINE_TIMER_MSECS / TCP_TMR_INTERVAL)))
                dhcp_fine_tmr();
        /* The dhcp_coarse_tmr() is called from dhcp_tmr_cb() since we need
           to call this also when link is down */
        if (!(priv->tick % (IP_TMR_INTERVAL / TCP_TMR_INTERVAL)))
                ip_reass_tmr();
#if LWIP_DNS == 1
        if (!(priv->tick % (DNS_TMR_INTERVAL / TCP_TMR_INTERVAL)))
                dns_tmr();
#endif
}

static void
dhcp_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        owl_unused(tmo);
        owl_unused(ctx);

        dhcp_coarse_tmr();

        if (! owl_bit(priv->flags, FLAG_DHCP))
                return;
        
        if (netif_is_up(priv->netif))
                return;
        
        if (owltmr_intvl_passed(priv->linkup_time, 
                                owltmr_get_ms(),
                                DHCP_ADDR_TMO)) {
                owl_dbg("Restarting DHCP");
                dhcp_renew(priv->netif);
                priv->linkup_time = owltmr_get_ms();
        }
}

static void
wifi_tmr_cb(struct owltmr_tmo *tmo, void *ctx)
{
        owl_unused(tmo); /* unused */
        owl_unused(ctx); /* unused */

        wlif_poll(priv->netif);
}
#endif /* !OWL_CONFIG_SOCKET_BSD */

#if OWL_CONFIG_HIBERNATE
struct hibernate_ctx {
        uint8_t flags;
        struct ip_addr ipaddr;
        struct ip_addr netmask;
        struct ip_addr gw;
        struct ip_addr dns;
};

int
owlip_serialize(void *data, int len)
{
        struct hibernate_ctx *p = data;
        
        owl_info("data:%p len:%d", data, len);

        if (data == NULL)
                return OWL_ERR_PARAM;

        if (len < sizeof(*p)) {
                owl_err("not enough room, need:%d len:%d", sizeof(*p), len);
                return OWL_ERR_SIZE;
        }

        p->flags = 0;
        if (owl_bit(priv->flags, FLAG_DHCP_INFORM)) {
                owl_bitset(p->flags, FLAG_DHCP_INFORM);
                ip_addr_set(&p->ipaddr, &priv->ipaddr);
                ip_addr_set(&p->netmask, &priv->netmask);
                ip_addr_set(&p->gw, &priv->gw);
                ip_addr_set(&p->dns, &priv->dns);
                
                owl_dbg("ipaddr:%s", ip_ntoa(&p->ipaddr));
                owl_dbg("netmask:%s", ip_ntoa(&p->netmask));
                owl_dbg("gw:%s", ip_ntoa(&p->gw));
                owl_dbg("dns:%s", ip_ntoa(&p->dns));
        }
        
        return sizeof(*p);
}

int
owlip_deserialize(const void *data, int len)
{
        const struct hibernate_ctx *p;
        
        owl_info("data:%p len:%d", data, len);
        
        if (data == NULL)
                return OWL_ERR_PARAM;

        /* Deserialize */
        p = data;
        if (len < sizeof(*p)) {
                owl_err("not enough data, need:%d len:%d", sizeof(*p), len);
                return OWL_ERR_SIZE;
        }

        if (owl_bit(p->flags, FLAG_DHCP_INFORM)) {
                owl_bitset(priv->flags, FLAG_DHCP_INFORM);
                ip_addr_set(&priv->ipaddr, &p->ipaddr);
                ip_addr_set(&priv->netmask, &p->netmask);
                ip_addr_set(&priv->gw, &p->gw);
                ip_addr_set(&priv->dns, &p->dns);

                owl_dbg("ipaddr:%s", ip_ntoa(&priv->ipaddr));
                owl_dbg("netmask:%s", ip_ntoa(&priv->netmask));
                owl_dbg("gw:%s", ip_ntoa(&priv->gw));
                owl_dbg("dns:%s", ip_ntoa(&priv->dns));
        }
        
        return sizeof(*p);
}
#endif /* OWL_CONFIG_HIBERNATE */
