/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#ifndef WLP_SPB8XX_H
#define WLP_SPB8XX_H

#include <owl_cc.h>

#include <wlp_ipaddr.h>
#include <wlp_defs.h>
#include <wlp_header.h>

#if defined(WLP_CONFIG_TRAMPOLINE)

int _wlp_init(int interface,
              int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
              void (*select_cb)(void *ctx, int select),
              void (*sdio_cmd_cb) (void *bus, uint8_t idx, uint32_t arg, 
                               uint8_t flags, uint32_t *rsp,
                               const void *data, int len),
              void *ctx);
void _wlp_shutdown(void);
int _wlp_set_baudrate(uint32_t baudrate, uint8_t rtscts);
int _wlp_fw_upgrade(uint8_t method, struct ip_addr *addr, uint16_t port,
                    const char *path);
int _wlp_set_fw_upgrade_cb(void (*status_cb)(void *ctx, uint8_t success),
                           void *ctx);
int _wlp_get_fw_version(uint32_t *code, char *build, int build_len,
                        char *hw_rev, int hw_rev_len);
int _wlp_get_fw_caps(uint8_t *cap_flags);
int _wlp_set_wakeup_cb(void (*wakeupf)(void *ctx, int enable),
                      void *ctx);
int _wlp_set_heartbeat(uint32_t ms_period);
int _wlp_set_led(int enable);
int _wlp_set_poll_period(uint32_t ms_period);
int _wlp_loopback(char *in, char *out, int len);

#else /* defined(WLP_CONFIG_TRAMPOLINE) */

#define wlp_init                _wlp_init
#define wlp_shutdown            _wlp_shutdown
#define wlp_set_baudrate        _wlp_set_baudrate
#define wlp_reset               _wlp_reset
#define wlp_fw_upgrade          _wlp_fw_upgrade
#define wlp_set_fw_upgrade_cb   _wlp_set_fw_upgrade_cb
#define wlp_get_fw_version      _wlp_get_fw_version
#define wlp_get_fw_caps         _wlp_get_fw_caps
#define wlp_set_wakeup_cb       _wlp_set_wakeup_cb
#define wlp_set_heartbeat       _wlp_set_heartbeat
#define wlp_set_led             _wlp_set_led
#define wlp_set_poll_period     _wlp_set_poll_period
#define wlp_loopback            _wlp_loopback

#endif /* defined(WLP_CONFIG_TRAMPOLINE) */


/*! \defgroup wlp_module Module specific functions.
 *
 * @{
 *
 */ 


/**
 * \brief Initialize the wifi device.
 *
 * Initialize the wlp api and configure the uart settings on the wifi device.
 * Must be called before any other function in the wlp api.
 *
 * wlp_init() will send a 'w' character, a 'w' will be received once the wifi
 * device is ready. The 'w' chars must be transmitted and received at
 * 115200 8N1.
 *
 * Note again that the uart which the callback parameters readf and writef
 * uses must be configured for 115200 8N1 before calling wlp_init().
 *
 * Allow the wifi device up to 200 ms to boot before calling wlp_init() after
 * power up.
 *
 * The sequence on the host becomes
 *  1) Power on host and wifi device
 *  2) Wait 200 ms
 *  3) Call wlp_init() with host uart configured as 115200 8N1
 *  4) Done, now the wifi device is ready to use.
 *
 * @param interface selects the host communication interface. This can be either
 *              WLP_INTERFACE_UART or WLP_INTERFACE_SPI.
 * @param exchange_cb is a pointer to a function which reads and writes chars
 *              from the bus which the wifi device is connected. If tx is not
 *              NULL, the buffer pointed to by the tx parameter should be
 *              written to the bus. If rx is not NULL, data should be read
 *              from the bus and stored in the buffer pointed to by the rx
 *              parameter. If both tx and rx are not NULL, a write should be
 *              followed by a read. Note that exchange_cb should block until
 *              the number of bytes specified in the len parameter are
 *              read/written.
 * @param select_cb will be invoked before and after any write-read transaction.
 *              For UART interfaces, this function could e.g. disable/enable
 *              interrupts on the host to ensure the UART data flow. For SPI
 *              interfaces, this could assert/desassert the chip select signal
 *              connected to the device. If the select parameter is 1, the
 *              transaction will start when this function has returned. If the
 *              select parameter is 0, the transaction is completed.
 * @param ctx is an opaque context pointer that will be passed back to readf,
 *              writef and selectf.
 * 
 * @return
 *   - 0 on success
 *
 */
int wlp_init(int interface,
             int (*exchange_cb)(void *ctx, const void *tx, void *rx, int len),
             void (*select_cb)(void *ctx, int select),
             void (*sdio_cmd_cb) (void *bus, uint8_t idx, uint32_t arg, 
                                  uint8_t flags, uint32_t *rsp,
                                  const void *data, int len),
             void *ctx);


/**
 * \brief Change the baudrate of the device.
 *
 * Send a uart configuration message which will configure the wifi device
 * according to the provided baudrate and rtscts parameters.
 *
 * Upon successful return, the wifi device will use the new uart configuration.
 * The host must change its uart settings accordingly before any other function
 * in wlp_api is called.
 *
 * The sequence on the host becomes
 *  1) Call wlp_set_baudrate() to reconfigure the device baudrate.
 *  2) Wait 250 ms after wlp_set_baudrate() has returned to allow the device to
 *     reconfigure its uart.
 *  3) Reconfigure host uart baudrate.
 *  4) Done, now the wifi device is ready to use with the new baudrate setting.
 *
 * The device supports the following baudrates:
 *
 * \verbatim
 *
 *    Baudrate | Error (%)
 *    --------------------
 *       9600  | 0
 *      19200  | 0
 *      38400  | 0
 *      57600  | 0.08
 *     115200  | 0.16
 *     230400  | 0.16
 *     460800  | 0.16
 *     921600  | 0.16
 *    1000000  | 0
 *    1500000  | 0
 *    2000000  | 0
 *    3000000  | 0
 *    4000000  | 0
 *
 * \endverbatim
 *
 * Note that the actual baudrate used might differ from the desired baudrate.
 * This is due to the fractional baud rate generation. It is recommended to
 * keep the sum of the transmitter and receiver error less than 2.00 %. As shown
 * by the table above, the SPB800 has a very small fractional error on the
 * baudrates.
 * 
 * @param baudrate is the baudrate which the wifi device should switch to.
 * @param rtcts is the hardware handshake configuration which the wifi device
 *              should switch to. Set to 1 to enabled hardware handshaking, and
 *              0 to disable hardware handshaking. Any hardware handshake will
 *              use the RTS and CTS pins.
 * @param readf is a pointer to a function which reads chars from the bus which
 *              the wifi device is connected to and stores them in the location
 *              pointed to by the data paramter. Note that readf should block
 *              until the number of bytes specified in the len parameter are
 *              read.
 * @param writef should read chars from the location pointed to by the data
 *              paramter send them to the bus which the wifi device is connected
 *              to. Note that writef should block until the number of bytes
 *              specified in the len parameter are written.
 * @param ctx is an opaque context pointer that will be passed back to readf
 *              and writef.
 *
 * @return
 *   - 0 on success
 *   - WLP_ERR_PARAM if invalid baudrate or rtscts parameters.
 *
 */
int wlp_set_baudrate(uint32_t baudrate, uint8_t rtscts);

/**
 * \brief Prepare the device for a firmware upgrade operation.
 * 
 * @param method Specify how to retrieve the firmware image.
 *               The only currently supported method is 
 *               WLP_UPGRADE_METHOD_HTTP_POST.
 * @param addr IP address. This parameter is ignored with method 
 *             WLP_UPGRADE_METHOD_HTTP_POST.
 * @param port Port number.
 *             With WLP_UPGRADE_METHOD_HTTP_POST the \a port parameter
 *             specifies the port on which the upgrade web page will
 *             be presented.
 * @param path Path. This parameter is ignored with method
 *             WLP_UPGRADE_METHOD_HTTP_POST.
 * @return
 *     - 0 on success
 *     - WLP_ERR_PARAM
 */
int wlp_fw_upgrade(uint8_t method, 
                   struct ip_addr *addr, 
                   uint16_t port,
                   const char *path);

/**
 * \brief Register a firmware upgrade notification callback.
 * 
 * If wlp_fw_upgrade() has been called then the callback registered
 * with this call will be invoked to signal success of failure of
 * the upgrade process.
 *
 * @param status_cb Callback to be invoked when the upgrade process has
 *                  been completed. \a success will be 1 if the upgrade
 *                  was successful or 0 if the upgrade failed.
 * @return
 *     - 0 on success
 */
int wlp_set_fw_upgrade_cb(void (*status_cb)(void *ctx, uint8_t success),
                          void *ctx);


/**
 * \brief Get the firmware version.
 *
 * @param code will contain the fw version code upon successful return. The
 *              version can be compared to the driver version (WLP_VERSION_CODE)
 *              using the WLP_VERSION(), WLP_MAJOR_VERSION(),
 *              WLP_MINOR_VERSION() and the WLP_PATCH_VERSION() macros.
 *
 * @param build will hold the build information upon successful return. The
 *              build information is a string of information for the particular
 *              fw build used on the device. No more than \param len characters,
 *              including null termination, will be copied to the string pointed
 *              to by build. This parameter can be NULL.
 *
 * @param build_len   is the maximum amount of chars, including null termination,
 *              that will be copied to \param build.
 *
 * @param hw_rev will hold the hardware revision string upon successful
 *              return. The hardware revision string describes the particular
 *              PCB revision and model of the pico server device. \a hw_rev
 *              has the format "<rev>:<board>", or is an empty string if
 *              the revision information is not available on the current
 *              pico device. No more than \param len characters, including
 *              null termination, will be copied to the string. This parameter
 *              can be NULL.
 *
 * @param hw_rev_len   is the maximum amount of chars, including null termination,
 *              that will be copied to \param hw_rev.
 * @return
 *   - 0 on success
 *   - WLP_FAILURE if any failure occurs
 *
 *
 */
int wlp_get_fw_version(uint32_t *code, 
                       char *build, 
                       int build_len,
                       char *hw_rev, 
                       int hw_rev_len);

/**
 * \brief Get WiFi device capabilities.
 *
 * Some functionality is dependent on the capabilities of the
 * WiFi device that wlp_api interfaces with. \a wlp_get_fw_caps()
 * retrieves a flag field describing the capabilities of the
 * attached WiFi device.
 *
 * The following capabilities are supported:
 * 
 * - WLP_CAP_HA_PIN
 *
 *   The device supports a Host Attention pin. The presence of the
 *   Host Attention pin allows the WiFi device to alert the host when
 *   there is pending activity and \a wlp_poll() should be
 *   called. This allows the host to enter power save states if it can
 *   be woken up when the host attention pin is asserted. It also
 *   allows the WiFi device to spend longer times in power save state
 *   (since calling wlp_poll() will wake it up). The host attention
 *   pin is active low so that a low voltage on the pin means that the
 *   host should call \a wlp_poll() repeatedly until the host
 *   attention pin voltage changes to high. If the WLP_CAP_HA_PIN is
 *   not present in the \a cap_flags field then the host attention pin
 *   will not be manipulated by the WiFi device and wlp_poll() should
 *   be called as often as is practicable. The host cannot sleep for
 *   long periods if this is the case. See the \ref wlp_ps section for
 *   a further discussion on how to use this flag.
 *
 * - WLP_CAP_WAKEUP_PIN
 *
 *   The device supports a Wakeup pin. The presence of the Wakeup pin
 *   allows the WiFi device to enter deeper power save modes where
 *   the Wakeup pin is used to wake it up from sleep.
 *   

 * @param cap_flags Contains a bitmask of the supported
 *        capabilities flags on output.
 * @return 
 *   -  0 on success
 */
int wlp_get_fw_caps(uint8_t *cap_flags);


/*! \defgroup wlp_ps Functions to manage power consumption
 *
 * These functions configure features that affect power consumption.
 * Achieving low power operation is always a tradeoff between some
 * aspect of performance (or an entire feature) and the power
 * consumption of the device. Beware that these functions can make the
 * device difficult to reach. It is recommended that you verify the
 * functionality of your application before attempting to reduce the
 * power consumption with these functions.
 *

In order to reduce the power consumption of the whole system
there are three different domains to consider.

\li \ref radios
\li \ref wifi_device_ps
\li \ref host_ps

\section radios WiFi Radio Power Save

The WiFi radio power save modes are controlled by \a wlp_conf_ps()
and \a wlp_set_ps().  The WiFi will, if so commanded by \a
wlp_set_ps(), sleep whenever possible and will only wake up and turn
on its receiver and transmitter when necessary.

\b Performance \b impact : Turning on WiFi power save will increase the
average receive latency since the WiFi only gets notifications of
pending packets at most once every beacon interval (which is usually
100ms).

\b Example : Enabling the WiFi power save mode

\code
if (wlp_conf_ps(0,    // use_ps_poll 
                200,  // ps_traffic_timeout 
                2000, // ps_delay 
                1,    // rx_all_dtim 
                1     // listen_interval
            ) < 0) {
        handle_error();
        return;
}
// Enable WiFi power save
wlp_set_ps(1);
\endcode

\section wifi_device_ps WiFi Device Sleep


<h2>Using the WAKEUP pin</h2>

The WiFi device can only sleep if it can be woken up by the host when
the host wants to send a command or poll events. A requirement for the
WiFi device to enter sleep mode is the presence of the WAKEUP pin
which is used to wake the device whenever the host wants to
communicate with it. The WAKEUP pin is not necessary for correct
operation but the presence of the pin will cause the device to consume
significantly less power. 

\b Performance \b impact : A slight delay is introduced in the communication
with the device.

\b Pin \b usage

 1. Connect the WAKEUP pin on the WiFi device to a pin on the host
    that the application can control. The WAKEUP pin is an output pin
    on the host, active low, and has a pull-up resistor on the WiFi
    device.

 2. Write a function wakeup_cb(void *ctx, int enable) that sets 
    the WAKEUP pin if enable is 0 and clears the WAKEUP pin if enable is 1.
    The pin is active low which is why the logic may seem inverted.

 3. Register the wakeup_cb() function with the wlp_api library using
    \a wlp_set_wakeup_cb(). The library will then automatically use the
    wakeup_cb() function to wake up the WiFi device whenever necessary.

\b Example : Supporting the WAKEUP pin

\code
static void wakeup_cb(void *ctx, int enable) {
        if (enable) {
                // Signal wakeup (active low)
                owlgpio_clr(OWL_BOARD_SPB800P_WAKEUP_PIN);
                // Wait for the device to wake up
                owlboard_mdelay(30);
        }
        else {
                // Reset wakeup (active low)
                owlgpio_set(OWL_BOARD_SPB800P_WAKEUP_PIN);
        }
}

// Register a WAKEUP pin control callback.
wlp_set_wakeup_cb(wakeup_cb, sh);
\endcode

<h2>Disabling LEDs</h2>

If the WiFi device has LEDs those can be disabled to conserve power.

\b Performance \b impact : None.

\b Example : Disable LEDs

\code
wlp_set_led(0);
\endcode

<h2>Reducing the WiFi Device Internal Polling Interval</h2>

The IP stack on the WiFi device must be polled periodically. For TCP
compliance the stack must be polled at least once every 250ms. In
situations where minimum power consumption is desired and TCP is not
used the polling interval can be increased.

\b Performance \b impact : DHCP and ARP timeouts are adversely affected. TCP
functionality may be impaired. The poll period should not be increased
beyond 250ms unless the traffic patterns are well known and known to
work with the decreased timer resolution.

\b Example : Set the polling interval to 5s

\code
wlp_set_poll_period(5000);
\endcode

\section host_ps Host Sleep

The host must normally poll the WiFi device to get notifications of
asynchronous events (such as traffic pending from the WiFi).  This
requires the host to wake up periodically for the polling
operation. To allow the host to sleep longer a HOST_ATTENTION pin is
supported on some hardware revisions. If the WiFi device supports the
HOST_ATTENTION pin it will signal on this pin whenever events are
pending. The host only has to poll the WiFi device when the
HOST_ATTENTION pin is signalled. This allows the host to sleep
whenever the pin is not signalled.

\b Performance \b impact : None.

<h2>Using the HOST_ATTENTION pin</h2>

 1. Connect the HOST_ATTENTION pin on the WiFi device to a pin on the host that
    the application can control. The HOST_ATTENTION pin is an input pin
    on the host, active low, and should have a pull-up resistor on the
    host.

 2. Check if the WiFi device supports the HOST_ATTENTION pin by
    calling wlp_get_fw_caps(&caps) and checking if the flag WLP_CAP_HA_PIN
    is set in caps.

 3. Make the application check the status of the HOST_ATTENTION pin as
    frequently as possible. Whenever the HOST_ATTENTION pin is low
    then \a wlp_poll() should be called until the HOST_ATTENTION pin goes
    high.  This is different from the case when the HOST_ATTENTION pin
    is unavailable/unused (such as in legacy code, or when the
    WLP_CAP_HA_PIN is not set in the fw capabilities flag field), when
    the \a wlp_poll() function should be called unconditionally.

\b Example : Supporting the HOST_ATTENTION pin

\code
static int ha_pin_is_signalled(void)
{
        if (owlgpio_is_low(OWL_BOARD_SPB800P_HOST_ATTENTION_PIN))
                return 1;
        return 0;
}

uint8_t fw_caps;
wlp_get_fw_caps(&fw_caps);

for (;;) {
        ...
        if (fw_caps & WLP_CAP_HA_PIN) {
                // If the host attention pin is present we
                // only poll if the pin is asserted (active low).
                // When the pin is asserted there's an event
                // pending from the SPB800 module.
                if (ha_pin_is_signalled())
                        wlp_poll();
        }
        else {
                // Without the host attention pin we need to
                // poll the SPB800 periodically. This method will
                // prevent the SPB800 from using power save
                // level 1.
                wlp_poll();
        }
        ...
}                
\endcode

 * @{
 *
 */ 


/**
 * \brief Register a callback that will be called to wakeup the wifi device.
 *
 * The callback will be invoked to control a wakeup-pin if one is
 * present on the current host platform. The wakeup-pin, if present,
 * is connected to the wifi device and is required if the wifi device
 * should be able to operate in the lowest power mode (power save
 * level 1). If the \a wakeupf() function is called with enable set to
 * 1 the wakeup-pin must be pulled low and if enable is set to 0 the
 * wakeup-pin must be pulled high. If the \a wakeupf() callback is
 * not provided then the wifi device can only operate in power save
 * level 2.
 *
 * @param wakeupf Wakeup-pin control function.
 * @param ctx is an opaque context pointer that will be passed back to readf,
 *              writef and selectf.
 * 
 * @return
 *   - 0 on success
 *
 */
int wlp_set_wakeup_cb(void (*wakeupf)(void *ctx, int enable),
                      void *ctx);

/** 
 * \brief Configure the link heartbeat.
 *
 * Some access points disconnects stations that have been idle (that
 * have not sent any data) for some period of time. They may
 * disconnect the station explicitly or they may silently drop the
 * connection.  This can cause problems if the device is a passive
 * listener (waiting for data that may never come because the
 * connection has been silently dropped). The link heartbeat can be
 * used to work around this limitation.
 *
 * If the heartbeat \a period is > 0 then a NULL data frame will be
 * sent to the access point at the specified \a period whenever a WiFi
 * connection has been established.
 *
 * Bear in mind that a low value for \a period will increase the power
 * consumption of the device. A reasonable value is 60000 or 120000
 * (sending heartbeats every 1 or 2 minutes).
 *
 * This function does not do anything in the AP-mode.
 *
 * The default heartbeat period is 60s.
 *
 * @param period The heartbeat period in milliseconds. 0 will disable
 *               the heartbeat functionality.
 * @return
 *   - 0 on success
 */
int wlp_set_heartbeat(uint32_t ms_period);

/**
 * \brief Enable/disable LEDs on the device.
 *
 * DEPRECATED
 * 
 * The 2 LEDs that light when the WiFi link and the IP link is up can
 * be disabled to conserve power.
 *
 * This function does not control the state of the LEDs (it cannot be
 * used to light the LEDs). If called with \a enable set to 0 it will
 * disable the LEDs completely. If called with \a enable set to 1 the
 * LEDs will be controlled by the internal link control logic (which
 * is the default behaviour).
 *
 * @param enable 1 to turn on the LED link indicator function. 0 to 
 *               turn LEDs completely off.
 * @return 
 *   - 0 on success
 */
int wlp_set_led(int enable);

/**
 * \brief Configure the device IP-stack poll frequency.
 *
 * DEPRECATED
 *
 * The device must wake up periodically to run timers required by
 * the IP stack. The wake up frequency affects power consumption.  To
 * provide a IETF-compliant TCP connection the poll period must be
 * 250ms or less. If no TCP connection is active, or if only UDP/RAW
 * data is handled, then it may be possible to increase the poll
 * period to a couple of seconds to reduce power consumption.
 *
 * Be aware that long polling periods can have large and
 * hard-to-predict consequences on network connectivity at the IP
 * layer and above. 
 *
 * This function only controls the polling frequency of the IP stack
 * in the device itself. It is unrelated to the \a wlp_poll() function
 * which is running on the host, not on the device.
 *
 * Note that when the wakeup-pin is supported (see \a
 * wlp_set_wakeup_cb()) and asserted then the device is kept awake (and
 * continuously checking pending timers) irrespective of the poll
 * period. The poll period is only used when the device actually
 * sleeps.
 *
 * @param ms_period Poll period in ms.
 * @return
 *   - 0 on success
 */

int wlp_set_poll_period(uint32_t ms_period);

int wlp_loopback(char *in, char *out, int len);

/*! @} */

#endif /* WLP_SPB8XX_H */
