/*
 * Copyright (C) 2012, H&D Wireless AB All rights reserved.
 *
 * The license to use this software in whole and in part and to
 * redistribute it in any form follows with the WiFi HW module from H&D
 * Wireless and is granted under the following restrictions:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of H&D Wireless AB may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * 4. The software may only be used together with hardware from H&D
 *    Wireless all other use is prohibited.
 *
 * 5. The license to use and redistribute the software is granted
 *    together with the purchase of a hardware platform on a one to one
 *    basis.
 *
 * 6. The binary code may not be reversed engineered or by other means
 *    copied to circumvent this license.
 * 
 * THIS SOFTWARE IS PROVIDED BY H&D WIRELESS AB ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT
 * SHALL HD WIRELESS AB BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For more information regarding this software license Contact H&D
 * Wireless AB (support@hd-wireless.se).
 *
 */

#if defined(WLP_DEBUG)
#define OWL_TRACE OWL_TRACE_ERR
#define OWL_FILE "wlp_spi"
#include <owl/core/owl_debug.h>
#include <owl/core/owl_err.h>
# define os_vdbg owl_vdbg
# define os_dbg  owl_dbg
# define os_info owl_info
# define os_err  owl_err
# define os_hexdump owl_hexdump
#else
# define os_vdbg(...)
# define os_dbg(...)
# define os_info(...)
# define os_err(...)
# define os_hexdump(...)
#endif /* WLP_DEBUG */


#if defined(WLP_LIBC)
# include <string.h>
# include <stdlib.h>
# define os_zalloc(x) calloc(x, 1) 
# define os_free      free 
# define os_strcpy    strcpy 
# define os_strncpy   strncpy 
# define os_strlen    strlen 
# define os_memcpy    memcpy 
# define os_memset    memset 
#else
# include <owl/core/owl_os.h>
#endif

#include <wlp_spi.h>
#include <wlp_defs.h>
#include <wlp_api.h>


#define bswap16(x)                        \
        (((((uint16_t) (x)) << 8) & 0xff00) |  \
         ((((uint16_t) (x)) >> 8) & 0xff))

# define bswap32(x) (((((uint32_t) (x)) << 24) & 0xff000000) |    \
                         ((((uint32_t) (x)) << 8) & 0xff0000) |       \
                         ((((uint32_t) (x)) >> 8) & 0xff00) |         \
                         ((((uint32_t) (x)) >> 24) & 0xff))

#if BYTE_ORDER == LITTLE_ENDIAN
#define cpu_to_le16(x)     (x)
#define cpu_to_le32(x)     (x)
#define cpu_to_be16(x)     bswap16(x)
#define cpu_to_be32(x)     bswap32(x)
#elif BYTE_ORDER == BIG_ENDIAN
#define le16_to_cpu(x)     bswap16(x)
#define le32_to_cpu(x)     bswap32(x)
#define be16_to_cpu(x)     (x)
#define be32_to_cpu(x)     (x)
#endif

#define MAX_BLOCK_LEN        511
#define RX_MIN_PKT_LEN       32
#define MAX_PKT_LEN          1536
#define SPI_HISPEED_CMD      0x8801e601

struct spi_dev {        
        int (*exchange_f)(void *ctx, const void *tx, void *rx, int len);
        void (*select_f)(void *ctx, int select);
        void *ctx;

	/* Receive params */
	uint32_t remain;
	uint32_t remain_idx;
	uint32_t total_block_len;
	uint32_t recvd;

        /* FIX ME */
        uint8_t tmp_buf[1500];
        uint8_t more_data[32];
};

static struct spi_dev spi_ctx;
static struct spi_dev *priv = NULL;

static int spi_cmd0(void);
static void spi_write_cmd52(uint32_t function);
static uint16_t spi_write_cmd53(const void *buf, int len);
static uint16_t spi_read_cmd53(void *buf, int len);
int wlp_spi_rx(void *buf, int len);
int wlp_spi_tx(const void *buf, int len);

int 
wlp_spi_init(int (*exchange_cb)
	     (void *ctx, const void *tx, void *rx, int len),
	     void (*select_cb)(void *ctx, int select),
             int *flags,
	     void *ctx)
{
        int err;

        priv = &spi_ctx;

        priv->exchange_f = exchange_cb;
	priv->select_f = select_cb;
        priv->ctx = ctx;

        /* Set device to SPI mode */
        if ((err = spi_cmd0()) < 0)
		return err;

        if (*flags & WLP_SPI_HISPEED)
                spi_write_cmd52(SPI_HISPEED_CMD);

        return 0;
}

int 
wlp_spi_tx(const void *buf, int len)
{
        uint32_t pos = 0;
        uint32_t tx_len;

	uint16_t *hdr;
	uint16_t aligned_size;

	os_memset(priv->tmp_buf, 0, sizeof(priv->tmp_buf));

	/* Align data block */
	aligned_size = len % 4;
	if (aligned_size > 0)
		aligned_size = 4 - aligned_size;
	
	aligned_size += len;

	/* Total size of packet including 2 byte len */
	hdr = (uint16_t *) &priv->tmp_buf[0];
	*hdr = (uint16_t) (aligned_size + 2);

	/* Real size of data */
	hdr = (uint16_t *) &priv->tmp_buf[2];
	*hdr = (uint16_t) len;

        os_memcpy(&priv->tmp_buf[4], buf, len);
	
	aligned_size += 4;

	while (aligned_size) {
                if (aligned_size > MAX_BLOCK_LEN)
                        tx_len = MAX_BLOCK_LEN;
                else
                        tx_len = aligned_size;
		
                pos += spi_write_cmd53(priv->tmp_buf + pos, tx_len);
		aligned_size -= tx_len;
        }
        return len;
}

int 
wlp_spi_rx(void *buf, int len)
{
        int pos = 0;
	int data_len = 0;
		
	/* If we have data remaining from last read, 
	   copy this to user buf */
	if (priv->remain) {
		uint32_t remain_idx = priv->remain_idx;
		uint32_t remain_len = priv->remain;
		
		if (len <= remain_len) {
                        os_memcpy(buf, priv->more_data + remain_idx, len);
			priv->remain -= len;
			return len;
		} else {
                        os_memcpy(buf, priv->more_data + remain_idx, remain_len);
			pos = priv->remain;
			priv->recvd += len;
			priv->remain = 0;
			priv->remain_idx = 0;
		}
	} else {
		priv->total_block_len = 0;
		priv->recvd = 0;

		/* Read 32 bytes */
		pos += spi_read_cmd53(priv->tmp_buf, RX_MIN_PKT_LEN);
       
		/* Get real data len at offset 2 (SDIO size header) */ 
		data_len = cpu_to_le16(((uint16_t*) priv->tmp_buf)[1]);
		
		/* offset to real data */
		pos -= 4;

		/* We got all data in first read */
		if (len <= (RX_MIN_PKT_LEN - 4) && len >= data_len) {
                        os_memcpy(buf, &priv->tmp_buf[4], len);
			return len;

		} else if (len == data_len) {
                        os_memcpy(buf, &priv->tmp_buf[4], RX_MIN_PKT_LEN - 4);
			
		} else if (len < data_len) {
			uint32_t tmp_len;
			priv->total_block_len = data_len;
			priv->recvd = len;
				
			/* We need to read 32 bytes at a time. See if 
			   we got more data than expected. If so, 
			   store data in buf so we can append
			   this to user buf at next read */
                        os_memcpy(buf, &priv->tmp_buf[4], len);
			
			if (data_len <= (RX_MIN_PKT_LEN - 4))
				tmp_len = data_len;
			else
				tmp_len = RX_MIN_PKT_LEN - 4;

                        os_memcpy(priv->more_data,
				      &priv->tmp_buf[4 + len], 
				      tmp_len - len);	      
			
			priv->remain = tmp_len - len;
			priv->remain_idx = 0;
			return len;
		}
	} 
      	/* More data? */
        while (pos < len) {
                uint32_t rx_len;

                /* Read remaining bytes, limit to sdio block size */
                rx_len = len - pos;
                if(rx_len > MAX_BLOCK_LEN) 
                        rx_len = MAX_BLOCK_LEN;
        
                pos += spi_read_cmd53((uint8_t *) buf + pos, rx_len);
	}
	/* Read the last padding bytes */
	if (priv->recvd < priv->total_block_len) {
	        int rest_len = priv->total_block_len - priv->recvd;
		rest_len = spi_read_cmd53((uint8_t *) priv->more_data, rest_len);
		priv->remain = rest_len;
	}
	
        return pos;
}

int
wlp_spi_exchange(void *ctx, const void *tx, void *rx, int len)
{
        (void) ctx;

        int err = 0;

        if (tx)
                err = wlp_spi_tx(tx, len);
        if (rx)
                err = wlp_spi_rx(rx, len);

	return err;
}

static int 
spi_cmd0(void)
{
        uint8_t zero[32] = { 0 };
        uint8_t cmd0[] = {
                0xff, 0xff,
                0x40, 
                0x00, 0x00, 0x00, 0x00, 
                0x01,
                0xff,
	};
	uint8_t rsp[17] = {0};

        priv->select_f(priv->ctx, 1);
        priv->exchange_f(priv->ctx, zero, NULL, sizeof(zero));
        priv->exchange_f(priv->ctx, cmd0, NULL, sizeof(cmd0));
        priv->exchange_f(priv->ctx, NULL, rsp, sizeof(rsp));
	priv->select_f(priv->ctx, 0);
	
        if(rsp[0] != 1)
                return WLP_ERR_SYNC;

        return 0;
}

static void 
spi_write_cmd52(uint32_t function)
{
        uint8_t cmd52[] = {
                0xff, 0xff, 0xff,
                0x74, 
                0x00, 0x00, 0x00, 0x00, 
                0xff,
                0xff,
                0xff, 0xff, 0xff 
        };
	
	uint8_t rsp[13] = {0};

	cmd52[4] = (function >> 24) & 0xff;
	cmd52[5] = (function >> 16) & 0xff;
	cmd52[6] = (function >> 8) & 0xff;
	cmd52[7] = function & 0xff;

	priv->select_f(priv->ctx, 1);
        priv->exchange_f(priv->ctx, cmd52, rsp, sizeof(cmd52));
       	priv->select_f(priv->ctx, 0);
}

static uint16_t 
spi_write_cmd53(const void *buf, int len)
{
	uint8_t cmd53[] = {
		0xff, 0xff, 0xff,
		0x75, 0x90, 0x00,
		0x00, 0x00,      
		0x01,
		0xff,
		0xff, 0xff,
		0xff, 0xfe
	};
	
	uint8_t rsp[14] = {0};

	cmd53[6] = (uint8_t) (len >> 8);
	cmd53[7] = (uint8_t) (len & 0xff);
	
	priv->select_f(priv->ctx, 1);
	priv->exchange_f(priv->ctx, cmd53, rsp, sizeof(cmd53));
	priv->exchange_f(priv->ctx, buf, NULL, len + 2); // Padding 2 bytes
	priv->select_f(priv->ctx, 0);

	return len;
}

static uint16_t 
spi_read_cmd53(void *buf, int len)
{
	uint8_t cmd53[] = {
		0xff, 0xff, 0xff,
		0x75, 0x10, 0x00,
		0x00, 0x00,      
		0x01,
		0xff,
		0xff, 0xff,
		0xff, 0xfe
	};
	
	uint8_t rsp[14] = {0};

	cmd53[6] = (uint8_t) (len >> 8);
	cmd53[7] = (uint8_t) (len & 0xff);
	
	priv->select_f(priv->ctx, 1);
	priv->exchange_f(priv->ctx, cmd53, rsp, sizeof(cmd53));
	priv->exchange_f(priv->ctx, NULL, buf, len + 4); // Padding bytes
	priv->select_f(priv->ctx, 0);

	return (cmd53[6] << 8) | cmd53[7];
}
